﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.Win32;

namespace Typoworm
{
    public partial class Game : Form
    {
        private static Color WormColorNormal = Color.LightSeaGreen;
        private static Color WormColorCorrect = Color.WhiteSmoke;
        private static Color WormColorWrong = Color.Red;

        private enum GameState
        {
            Menu,
            Playing,
            End
        }

        private class GameChar
        {
            public char ch;
            public string ch_str;
            public float x;
            public float y;

            public GameChar(char _ch, float _x, float _y)
            {
                ch = _ch;
                ch_str = ch.ToString();
                x = _x;
                y = _y;
            }
        }

        private string[] m_word_list = 
        {
            "Typo",
            "Worm",
            "Cave",
            "Moss",
            "Rock",
            "Moist",
            "Dry",
            "Wet",
            "Gravity",
            "Faster",
            "Slower",
            "Fly",
            "Mole",
            "Gold",
            "Silver",
            "Bronze",
            "Platinum",
            "Mine",
            "Dirt",
            "Mine",
            "Beam",
            "Support",
            "Dig",
            "Crawl",
            "Soft",
            "Hard",
            "Granite",
            "Solid",
            "Sediment",
            "Shovel",
            "Truck",
            "Slither",
        };
        private List<GameChar> m_wall_chars = new List<GameChar>();
        private List<GameChar> m_bg1_chars = new List<GameChar>();
        private List<GameChar> m_bg2_chars = new List<GameChar>();
        private List<char> m_type_chars = new List<char>();
        private List<char> m_typed_chars = new List<char>();

        private Font m_bg_font;
        private Font m_text_font;
        private Font m_score_font;
        private Font m_highscore_font;

        private Random m_random = new Random();

        private GameState m_state;

        private float m_wall_size_percent;
        private float m_wall_pos_percent;
        private float m_wall_spawn_timer;

        private float m_bg_spawn_timer;

        private float m_wall_pos_delta;
        private int m_wall_pos_timer;

        private float m_worm_x;
        private float m_worm_y;
        private float m_worm_y_delta;
        private Color m_worm_color;

        private float m_end_timer;
        private float m_end_anim;

        private int m_score;
        private int m_highscore;

        public Game()
        {
            InitializeComponent();

            m_bg_font = new Font(Font.FontFamily, Font.Size * 6.0f, FontStyle.Bold | FontStyle.Italic);
            m_text_font = new Font(Font.FontFamily, Font.Size * 0.7f, FontStyle.Bold);
            m_score_font = new Font(Font.FontFamily, Font.Size * 0.5f, FontStyle.Bold);
            m_highscore_font = new Font(Font.FontFamily, m_score_font.Size * 0.8f, FontStyle.Bold);

            // Get the highscore
            {
                RegistryKey hklm = Registry.CurrentUser;
                hklm = hklm.CreateSubKey("Software\\XMunkki\\Typoworm");
                try { m_highscore = (int)hklm.GetValue("Highscore"); }
                catch { }
            }

            // Start a new game
            NewGame();
        }

        public void NewGame()
        {
            m_state = GameState.Menu;

            m_wall_size_percent = 0.0f;
            m_wall_pos_percent = 0.5f;
            m_wall_spawn_timer = 0.0f;

            m_bg_spawn_timer = 0.0f;

            m_wall_pos_delta = 0.0f;
            m_wall_pos_timer = 0;

            m_worm_x = 10;
            m_worm_y = (float)ClientRectangle.Height * 0.5f;
            m_worm_y_delta = 0.0f;
            m_worm_color = WormColorNormal;

            m_end_timer = 0.0f;
            m_end_anim = 0.0f;

            m_score = 0;

            m_wall_chars.Clear();
            m_bg1_chars.Clear();
            m_bg2_chars.Clear();
            m_typed_chars.Clear();
            m_type_chars.Clear();

            foreach (char c in "StartTypingToBegin")
                m_type_chars.Add(c);

            MoreTypeChars();
        }

        private void MoreTypeChars()
        {
            // Add new characters (as needed)
            while (m_type_chars.Count < 100)
            {
                string word = m_word_list[m_random.Next(m_word_list.Length)];
                foreach (char c in word)
                    m_type_chars.Add(c);
            }

            // Remove old typed characters
            while (m_typed_chars.Count > 100)
                m_typed_chars.RemoveAt(0);
        }

        private void Game_Paint(object sender, PaintEventArgs e)
        {
            Graphics g = e.Graphics;
            int w = ClientRectangle.Width;
            int h = ClientRectangle.Height;

            g.Clear(Color.Black);

            // Draw the background
            {
                Brush bg1_brush = new SolidBrush(Color.FromArgb(30, 30, 30));
                Brush bg2_brush = new SolidBrush(Color.FromArgb(50, 50, 50));

                foreach (GameChar ch in m_bg1_chars)
                {
                    SizeF size = g.MeasureString(ch.ch_str, m_bg_font);
                    g.DrawString(ch.ch_str, m_bg_font, bg1_brush, ch.x, ch.y - size.Height * 0.5f);
                }
                foreach (GameChar ch in m_bg2_chars)
                {
                    SizeF size = g.MeasureString(ch.ch_str, m_bg_font);
                    g.DrawString(ch.ch_str, m_bg_font, bg2_brush, ch.x, ch.y - size.Height * 0.5f);
                }
            }

            // Draw the walls
            foreach (GameChar ch in m_wall_chars)
            {
                //g.FillRectangle(Brushes.Red, ch.x - 3, ch.y - 3, 6, 6);

                SizeF size = g.MeasureString(ch.ch_str, Font);
                g.DrawString(ch.ch_str, Font, Brushes.White, ch.x - size.Width * 0.5f, ch.y - size.Height * 0.5f);
            }

            // Draw the worm
            g.FillEllipse(new SolidBrush(m_worm_color), m_worm_x - 3, m_worm_y - 3, 6, 6);

            // Draw the "explosion"
            if (m_state == GameState.End)
            {
                float expl_size2 = m_end_anim * 0.5f;
                g.FillEllipse(new SolidBrush(Color.FromArgb(60, 255, 50, 10)), m_worm_x - expl_size2, m_worm_y - expl_size2, m_end_anim, m_end_anim);
            }

            // Draw the text area
            {
                g.FillRectangle(new SolidBrush(Color.FromArgb(200, 0, 20, 100)), 0, h - 20, w, 20);

                float text_y = h - 20;
                float text_x = m_worm_x;

                // New chars
                for (int i = 0; i < m_type_chars.Count; ++i)
                {
                    if (text_x > w)
                        break;

                    // Get the color
                    float lerp_factor = (float)i * 0.3f;
                    if (lerp_factor > 1.0f)
                        lerp_factor = 1.0f;
                    Color col = LerpColor(Color.White, Color.LightGreen, lerp_factor);

                    // Write the string
                    string ch_str = m_type_chars[i].ToString();
                    g.DrawString(ch_str, m_text_font, new SolidBrush(col), text_x, text_y);

                    // Advance
                    SizeF size = g.MeasureString(ch_str, m_text_font);
                    text_x += size.Width * 0.8f;
                }

                // Old chars
                Brush old_text_brush = new SolidBrush(Color.Gray);

                text_x = m_worm_x;
                for (int i = m_typed_chars.Count - 1; i >= 0; --i)
                {
                    if (text_x < -10)
                        break;

                    // Advance
                    string ch_str = m_typed_chars[i].ToString();
                    SizeF size = g.MeasureString(ch_str, m_text_font);
                    text_x -= size.Width * 0.8f;

                    // Write the string
                    g.DrawString(ch_str, m_text_font, old_text_brush, text_x, text_y);
                }
            }

            // Draw the scores
            {
                string score_txt = m_score.ToString();
                string highscore_txt = m_highscore.ToString();

                g.DrawString(score_txt, m_score_font, Brushes.Black, 2, 2);
                g.DrawString(score_txt, m_score_font, Brushes.Red, 0, 0);

                SizeF size = g.MeasureString(score_txt, m_score_font);
                g.DrawString(highscore_txt, m_highscore_font, Brushes.Black, 2, size.Height * 0.7f + 2);
                g.DrawString(highscore_txt, m_highscore_font, Brushes.Brown, 0, size.Height * 0.7f);
            }

            // Draw the title
            if (m_state == GameState.Menu)
            {
                Font font2 = new Font(Font.FontFamily, Font.Size * 0.5f, FontStyle.Bold);

                WriteCentered(g, Font, Brushes.White, "Typoworm", w * 0.5f, h * 0.2f);
                WriteCentered(g, font2, Brushes.LightGray, "11th Ludumdare 48h entry", w * 0.5f, h * 0.3f);

                WriteCentered(g, font2, Brushes.DarkCyan, "Jere Sanisalo", w * 0.5f, h * 0.6f);
                WriteCentered(g, font2, Brushes.DarkCyan, "http://www.xmunkki.org/", w * 0.5f, h * 0.68f);
            }
        }

        private void WriteCentered(Graphics g, Font font, Brush brush, string str, float x, float y)
        {
            SizeF size = g.MeasureString(str, font);
            g.DrawString(str, font, brush, x - size.Width * 0.5f, y - size.Height * 0.5f);
        }

        private float Lerp(float v1, float v2, float t)
        {
            return v1 + t * (v2 - v1);
        }

        private Color LerpColor(Color c1, Color c2, float t)
        {
            return Color.FromArgb(
                (int)Lerp(c1.A, c2.A, t),
                (int)Lerp(c1.R, c2.R, t),
                (int)Lerp(c1.G, c2.G, t),
                (int)Lerp(c1.B, c2.B, t));
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            float delta_t = 1.0f / (float)timer.Interval;

            switch (m_state)
            {
                case GameState.Menu:
                    break;

                case GameState.Playing:
                    UpdateGame(delta_t);
                    break;

                case GameState.End:
                    UpdateEnd(delta_t);
                    break;
            }


            // Repaint
            Invalidate();
        }

        private void UpdateGame(float delta_t)
        {
            int w = ClientRectangle.Width;
            int h = ClientRectangle.Height;

            // Grow the walls
            if (m_wall_size_percent < 0.65f)
            {
                float grow_speed = 0.01f;
                if (m_wall_size_percent < 0.2f)
                    grow_speed = 0.1f;
                if (m_wall_size_percent < 0.3f)
                    grow_speed = 0.05f;
                if (m_wall_size_percent < 0.4f)
                    grow_speed = 0.02f;

                m_wall_size_percent += grow_speed * delta_t;
            }

            // Move the walls up/down
            if (m_wall_pos_timer <= 0)
            {
                float new_dest = (float)m_random.NextDouble();
                while (Math.Abs(new_dest - m_wall_pos_percent) > 0.4f)
                {
                    new_dest = (float)m_random.NextDouble();
                }

                float new_time = 0.2f + (float)m_random.NextDouble() * 1.0f;

                m_wall_pos_timer = (int)Math.Ceiling(new_time / delta_t);
                m_wall_pos_delta = (new_dest - m_wall_pos_percent) / (float)m_wall_pos_timer;
            }
            else
            {
                --m_wall_pos_timer;
                m_wall_pos_percent += m_wall_pos_delta;
                if (m_wall_pos_percent < 0.0f)
                    m_wall_pos_percent = 0.0f;
                if (m_wall_pos_percent > 1.0f)
                    m_wall_pos_percent = 1.0f;
            }

            // Move the worm closer to the screen
            if (m_worm_x < (w * 0.85f))
            {
                m_worm_x += 5.0f * delta_t;
            }

            // Move the worm up and down
            m_worm_y_delta += 100.0f * delta_t;
            m_worm_y += m_worm_y_delta * delta_t;

            // Update the worm color
            m_worm_color = LerpColor(m_worm_color, WormColorNormal, 0.1f);

            // Collision test for the worm
            bool worm_hit = false;

            if (m_worm_y <= 0)
            {
                m_worm_y = 0;
                m_worm_y_delta = 0;
                worm_hit = true;
            }
            if (m_worm_y > (h - 20))
            {
                m_worm_y = (h - 20);
                m_worm_y_delta = 0;
                worm_hit = true;
            }

            if (!worm_hit)
            {
                foreach (GameChar ch in m_wall_chars)
                {
                    float dx = m_worm_x - ch.x;
                    float dy = m_worm_y - ch.y;
                    float dist2 = dx * dx + dy * dy;
                    if (dist2 < (10 * 10))
                    {
                        worm_hit = true;
                    }
                }
            }

            if (worm_hit)
            {
                // Save the high score
                {
                    RegistryKey hklm = Registry.CurrentUser;
                    hklm = hklm.CreateSubKey("Software\\XMunkki\\Typoworm");
                    try { hklm.SetValue("Highscore", m_highscore); }
                    catch { }
                }

                // Change state
                m_state = GameState.End;
                return;
            }

            // Spawn new wall characters?
            string wall_characters = "#@$¤£&%";

            m_wall_spawn_timer -= delta_t;
            if (m_wall_spawn_timer < 0.0f)
            {
                m_wall_spawn_timer = 0.15f;

                float wall_size = m_wall_size_percent * (float)h;
                float wall_size_top = wall_size * m_wall_pos_percent;
                float wall_size_bottom = wall_size - wall_size_top;

                for (float y = wall_size_top; y > -10; y -= 15)
                {
                    char ch = wall_characters[m_random.Next(wall_characters.Length)];
                    m_wall_chars.Add(new GameChar(ch, w + 20, y));
                }
                for (float y = h - wall_size_bottom; y < (h + 10); y += 15)
                {
                    char ch = wall_characters[m_random.Next(wall_characters.Length)];
                    m_wall_chars.Add(new GameChar(ch, w + 20, y));
                }
            }

            // Spawn new bg characters
            m_bg_spawn_timer -= delta_t;
            if (m_bg_spawn_timer < 0.0f)
            {
                m_bg_spawn_timer = 2.0f;

                float y1 = (float)m_random.NextDouble() * h;
                float y2 = (float)m_random.NextDouble() * h;
                char ch1 = wall_characters[m_random.Next(wall_characters.Length)];
                char ch2 = wall_characters[m_random.Next(wall_characters.Length)];

                m_bg1_chars.Add(new GameChar(ch1, w, y1));
                m_bg2_chars.Add(new GameChar(ch2, w, y2));
            }


            // Move all characters
            {
                int idx = 0;
                while (idx < m_wall_chars.Count)
                {
                    if (!UpdateGameChar(m_wall_chars[idx], delta_t))
                        m_wall_chars.RemoveAt(idx);
                    else
                        ++idx;

                }
            }

            {
                int idx = 0;
                while (idx < m_bg1_chars.Count)
                {
                    UpdateGameChar(m_bg1_chars[idx], delta_t * 0.3f);
                    if (m_bg1_chars[idx].x < -200)
                        m_bg1_chars.RemoveAt(idx);
                    else
                        ++idx;

                }
            }

            {
                int idx = 0;
                while (idx < m_bg2_chars.Count)
                {
                    UpdateGameChar(m_bg2_chars[idx], delta_t * 0.6f);
                    if (m_bg2_chars[idx].x < -200)
                        m_bg2_chars.RemoveAt(idx);
                    else
                        ++idx;

                }
            }

            // Add score
            ++m_score;
            if (m_score > m_highscore)
                m_highscore = m_score;
        }

        private void UpdateEnd(float delta_t)
        {
            // End timer
            m_end_timer += delta_t;
            if (m_end_timer > 1.0f)
            {
                NewGame();
                return;
            }

            // End animation
            m_end_anim += delta_t * 1000.0f;
        }

        private bool UpdateGameChar(GameChar ch, float delta_t)
        {
            ch.x -= 50.0f * delta_t;
            return (ch.x > -20.0f);
        }

        private void Game_KeyPress(object sender, KeyPressEventArgs e)
        {
            char user_ch = char.ToUpper(e.KeyChar);
            char game_ch = char.ToUpper(m_type_chars[0]);

            if (m_state == GameState.End)
                return;

            // Correct char pressed?
            if (user_ch == game_ch)
            {
                // Consume the character
                m_typed_chars.Add(m_type_chars[0]);
                m_type_chars.RemoveAt(0);
                MoreTypeChars();

                // Move upwards
                m_worm_y_delta -= 30.0f;

                // Set the worm color
                m_worm_color = WormColorCorrect;

                // Set the game mode
                if (m_state == GameState.Menu)
                {
                    m_state = GameState.Playing;
                }
            }
            else
            {
                if (m_state == GameState.Playing)
                {
                    // Penalty: move downwards
                    m_worm_y_delta += 15.0f;

                    // Set the worm color
                    m_worm_color = WormColorWrong;
                }
            }
        }
    }
}
